"""
This is the entry point for main "kuyruk" executable command.
It implements the command line parsing for subcommands and configuration.

"""
from __future__ import absolute_import
import sys
import logging
import argparse
import pkg_resources

from kuyruk import __version__, importer, Kuyruk
from kuyruk.worker import Worker


logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(conflict_handler='resolve')

    # Add common options
    parser.add_argument(
        '-v', '--version', action='version', version=__version__)
    parser.add_argument(
        '-a', '--app', required=True, help='Path to the Kuyruk object')

    subparsers = parser.add_subparsers(help='sub-command name')

    # Parser for the "worker" sub-command
    parser_worker = subparsers.add_parser('worker', help='run a worker')
    parser_worker.set_defaults(func=run_worker)
    parser_worker.add_argument(
        '-q', '--queue', default='kuyruk', help='consume tasks from')
    parser_worker.add_argument(
        '-l', '--local', action="store_true",
        help='append hostname to the queue name')

    # Add additional subcommands from extensions.
    for entry_point in pkg_resources.iter_entry_points("kuyruk.commands"):
        command_func, help_text, customize_parser = entry_point.load()
        ext_parser = subparsers.add_parser(entry_point.name, help=help_text)
        ext_parser.set_defaults(func=command_func)
        if customize_parser:
            customize_parser(ext_parser)

    # Parse arguments
    args = parser.parse_args()

    # Import Kuyruk app
    sys.path.insert(0, '')
    app = importer.import_object_str(args.app)
    assert isinstance(app, Kuyruk)

    # Run the sub-command function
    args.func(app, args)


def run_worker(app, args):
    w = Worker(app, args)
    w.run()


if __name__ == '__main__':
    main()



import os
import signal
import logging
import unittest

from mock import patch

from tests import tasks
from tests.integration.util import run_kuyruk, wait_until, \
    not_running, get_pid, TIMEOUT, delete_queue, len_queue


logger = logging.getLogger(__name__)


class KuyrukTestCase(unittest.TestCase):
    """
    Tests here are mostly integration tests. They require a running
    RabbitMQ and Redis instances to pass succesfully.
    Just like the normal user they spawn a real worker process
    and make some assertion in their output.

    """
    def setUp(self):
        delete_queue('kuyruk')

    def test_simple_task(self):
        """Run a task on default queue"""
        tasks.echo('hello world')
        with run_kuyruk() as worker:
            worker.expect('Consumer started')
            worker.expect('hello world')
            worker.expect('Task is processed')

    def test_another_queue(self):
        """Run a task on another queue"""
        tasks.echo_another('hello another')
        with run_kuyruk(queue='another_queue') as worker:
            worker.expect('Consumer started')
            worker.expect('another_queue')
            worker.expect('hello another')
            worker.expect('Task is processed')

    def test_exception(self):
        """Errored task message is discarded"""
        tasks.raise_exception()
        with run_kuyruk() as worker:
            worker.expect('ZeroDivisionError')
        assert len_queue("kuyruk") == 0

    def test_retry(self):
        """Errored task is retried"""
        tasks.retry_task()
        with run_kuyruk() as worker:
            worker.expect('ZeroDivisionError')
            worker.expect('ZeroDivisionError')
        assert len_queue("kuyruk") == 0

    def test_cold_shutdown(self):
        """Cold shutdown stops stuck worker"""
        tasks.loop_forever()
        with run_kuyruk(terminate=False) as worker:
            worker.expect('looping forever')
            worker.send_signal(signal.SIGINT)
            worker.expect('Warm shutdown')
            worker.expect('Handled SIGINT')
            worker.send_signal(signal.SIGINT)
            worker.expect('Cold shutdown')
            worker.expect_exit(0)
            wait_until(not_running, timeout=TIMEOUT)

    def test_reject(self):
        """Rejected task is requeued"""
        tasks.rejecting_task()
        with run_kuyruk() as worker:
            worker.expect('Task is rejected')
            worker.expect('Task is rejected')
        assert len_queue("kuyruk") == 1

    @patch('tests.tasks.must_be_called')
    def test_before_after(self, presend_mock):
        """Run signal handlers"""
        tasks.task_with_signal_handlers('hello world')
        presend_mock.assert_called_once()
        with run_kuyruk() as worker:
            worker.expect('function2')
            worker.expect('hello world')
            worker.expect('function5')

    def test_max_run_time(self):
        """Timeout long running task"""
        run_time = tasks.sleeping_task.max_run_time + 0.1
        tasks.sleeping_task(run_time)
        with run_kuyruk() as worker:
            worker.expect('Timeout')

    def test_worker_sigquit(self):
        """Ack current message and exit"""
        tasks.loop_forever()
        with run_kuyruk(terminate=False) as worker:
            worker.expect('looping forever')
            pid = get_pid('kuyruk: worker')
            os.kill(pid, signal.SIGQUIT)
            worker.expect('Dropping current task')
            worker.expect('Exiting')
            worker.expect_exit(0)
        assert len_queue("kuyruk") == 0, worker.get_output()



"""
Contains sample tasks to be used in unit and integration tests.

"""
from __future__ import print_function
import sys
import string
import random
from time import sleep

from kuyruk import Kuyruk, Task
from kuyruk import signals
from kuyruk.config import Config


# Override defaults for testing
Config.WORKER_LOGGING_LEVEL = "debug"
Config.WORKER_MAX_LOAD = 999


kuyruk = Kuyruk()
# These functions below needs to be at module level in order that
# Kuyruk worker to determine their fully qualified name.


@kuyruk.task
def echo(message):
    print(message)
    must_be_called()


@kuyruk.task(queue='another_queue')
def echo_another(message):
    print(message)


@kuyruk.task
def flood():
    s = ''.join(random.choice(string.ascii_uppercase) for _ in xrange(70))
    while True:
        sys.stdout.write(s)
        sys.stdout.write('\n')
        sys.stdout.flush()
        sys.stderr.write(s)
        sys.stderr.write('\n')
        sys.stderr.flush()


@kuyruk.task
def raise_exception():
    return 1 / 0


@kuyruk.task(retry=1)
def retry_task():
    return 1 / 0


@kuyruk.task
def loop_forever():
    while 1:
        print('looping forever')
        sleep(1)


config_eager = Config()
config_eager.EAGER = True
kuyruk_eager = Kuyruk(config_eager)
@kuyruk_eager.task()
def eager_task():
    must_be_called()


@kuyruk.task
def rejecting_task():
    raise kuyruk.Reject


@kuyruk.task(max_run_time=1)
def sleeping_task(seconds):
    sleep(seconds)


# Another kuyruk instance for testing before and after task decorators
kuyruk2 = Kuyruk()


@kuyruk2.task
def task_with_signal_handlers(message):
    print(message)
    return 42


@signals.task_presend.connect_via(kuyruk2)
def function0(sender, task, args, kwargs, description):
    must_be_called()


@signals.task_prerun.connect_via(kuyruk2)
def function2(sender, task, args, kwargs):
    print('function2')
    print(sender, task, args, kwargs)
    assert sender is kuyruk2
    assert isinstance(task, Task)
    assert list(args) == ['hello world']
    assert kwargs == {}


@signals.task_postapply.connect_via(kuyruk2)
def function5(sender, task, args, kwargs):
    print('function5')


def must_be_called():
    """This function is patched in tests to see the caller is doing it's job.
    """
    pass



from __future__ import absolute_import, print_function
import os
import sys
import json
import errno
import socket
import signal
import logging
import logging.config
import threading
import traceback
import multiprocessing
from time import time, sleep

from setproctitle import setproctitle

from kuyruk import importer, signals
from kuyruk.task import get_queue_name
from kuyruk.exceptions import Reject, Discard

logger = logging.getLogger(__name__)


class Worker(object):
    """Consumes tasks from a queue and runs them.

    :param app: An instance of :class:`~kuyruk.Kuyruk`
    :param args: Command line arguments

    """
    def __init__(self, app, args):
        self.kuyruk = app

        if not args.queue:
            raise ValueError("empty queue name")
        self.queue = get_queue_name(args.queue, local=args.local)

        self.shutdown_pending = threading.Event()
        self._pause = False
        self._consuming = False
        self._current_message = None
        if self.config.WORKER_MAX_LOAD is None:
            self.config.WORKER_MAX_LOAD = multiprocessing.cpu_count()

        signals.worker_init.send(self.kuyruk, worker=self)

    @property
    def config(self):
        return self.kuyruk.config

    def run(self):
        """Runs the worker and consumes messages from RabbitMQ.
        Returns only after `warm_shutdown()` is called.

        """
        setproctitle("kuyruk: worker on %s" % self.queue)

        self._setup_logging()

        signal.signal(signal.SIGINT, self._handle_sigint)
        signal.signal(signal.SIGTERM, self._handle_sigterm)
        signal.signal(signal.SIGQUIT, self._handle_sigquit)
        signal.signal(signal.SIGUSR1, print_stack)  # for debugging

        for f in (self._watch_load, self._shutdown_timer):
            t = threading.Thread(target=f)
            t.daemon = True
            t.start()

        self._consume_messages()

        signals.worker_shutdown.send(self.kuyruk, worker=self)
        logger.debug("End run worker")

    def _setup_logging(self):
        if self.config.WORKER_LOGGING_CONFIG:
            logging.config.fileConfig(self.config.WORKER_LOGGING_CONFIG)
        else:
            logging.getLogger('rabbitpy').level = logging.WARNING
            level = getattr(logging, self.config.WORKER_LOGGING_LEVEL.upper())
            fmt = "%(levelname).1s " \
                  "%(name)s.%(funcName)s:%(lineno)d - %(message)s"
            logging.basicConfig(level=level, format=fmt)

    def _consume_messages(self):
        with self.kuyruk as k:
            with k.channel() as ch:
                ch.queue_declare(
                    queue=self.queue, durable=True, auto_delete=False)

                # Set prefetch count to 1. If we don't set this, RabbitMQ keeps
                # sending messages while we are already working on a message.
                ch.basic_qos(0, 1, False)

                consumer_tag = '%s@%s' % (os.getpid(), socket.gethostname())
                while not self.shutdown_pending.is_set():
                    # Consume or pause
                    if self._pause and self._consuming:
                        ch.basic_cancel(consumer_tag)
                        logger.info('Consumer cancelled')
                        self._consuming = False
                    elif not self._pause and not self._consuming:
                        ch.basic_consume(queue=self.queue,
                                         consumer_tag=consumer_tag,
                                         callback=self._message_callback)
                        logger.info('Consumer started')
                        self._consuming = True

                    try:
                        ch.connection.drain_events(timeout=0.1)
                    except socket.error as e:
                        if isinstance(e, socket.timeout):
                            pass
                        elif e.errno == errno.EINTR:
                            pass  # happens when the process receives a signal
                        else:
                            raise
        logger.debug("End run worker")

    def _message_callback(self, message):
        # Save current message being processed so
        # we will be able to send ACK when we receive SIGQUIT."""
        self._current_message = message
        try:
            self._process_message(message)
        finally:
            self._current_message = None

    def _process_message(self, message):
        """Processes the message received from the queue."""
        try:
            description = json.loads(message.body)
        except Exception:
            message.channel.basic_reject(message.delivery_tag, requeue=False)
            logger.error("Cannot decode message. Dropping.")
        else:
            logger.info("Processing task: %r", description)
            self._process_description(message, description)

    def _process_description(self, message, description):
        try:
            task = importer.import_object(description['module'],
                                          description['function'])
            args, kwargs = description['args'], description['kwargs']
        except Exception:
            logger.error('Cannot import task')
            exc_info = sys.exc_info()
            signals.worker_failure.send(self.kuyruk, description=description,
                                        exc_info=exc_info, worker=self)
            message.channel.basic_reject(message.delivery_tag, requeue=False)
        else:
            self._process_task(message, description, task, args, kwargs)

    def _process_task(self, message, description, task, args, kwargs):
        try:
            task.apply(*args, **kwargs)
        except Reject:
            logger.warning('Task is rejected')
            sleep(1)  # Prevent cpu burning
            message.channel.basic_reject(message.delivery_tag, requeue=True)
        except Discard:
            logger.warning('Task is discarded')
            message.channel.basic_reject(message.delivery_tag, requeue=False)
        except Exception:
            logger.error('Task raised an exception')
            exc_info = sys.exc_info()
            logger.error(''.join(traceback.format_exception(*exc_info)))
            signals.worker_failure.send(self.kuyruk, description=description,
                                        task=task, args=args, kwargs=kwargs,
                                        exc_info=exc_info, worker=self)
            message.channel.basic_reject(message.delivery_tag, requeue=False)
        else:
            logger.info('Task is successful')
            message.channel.basic_ack(message.delivery_tag)
        finally:
            logger.debug("Task is processed")

    def _watch_load(self):
        """Pause consuming messages if lood goes above the allowed limit."""
        while not self.shutdown_pending.is_set():
            load = os.getloadavg()[0]
            if load > self.config.WORKER_MAX_LOAD:
                if self._pause is False:
                    logger.warning(
                        'Load is above the treshold (%.2f/%s), '
                        'pausing consumer', load, self.config.WORKER_MAX_LOAD)
                    self._pause = True
            else:
                if self._pause is True:
                    logger.warning(
                        'Load is below the treshold (%.2f/%s), '
                        'resuming consumer', load, self.config.WORKER_MAX_LOAD)
                    self._pause = False
            sleep(1)

    def _shutdown_timer(self):
        """Counts down from MAX_WORKER_RUN_TIME. When it reaches zero sutdown
        gracefully.

        """
        if not self.config.WORKER_MAX_RUN_TIME:
            return

        started = time()
        while not self.shutdown_pending.is_set():
            passed = time() - started
            remaining = self.config.WORKER_MAX_RUN_TIME - passed
            if remaining > 0:
                sleep(remaining)
            else:
                logger.warning('Run time reached zero')
                self.warm_shutdown()
                break

    def warm_shutdown(self):
        """Exits after the current task is finished."""
        logger.warning("Warm shutdown")
        self.shutdown_pending.set()

    def cold_shutdown(self):
        """Exits immediately."""
        logger.warning("Cold shutdown")
        _exit(0)

    def _handle_sigint(self, signum, frame):
        """If running from terminal pressing Ctrl-C will initiate a warm
        shutdown. The second interrupt will do a cold shutdown.

        """
        logger.warning("Handling SIGINT")
        if sys.stdin.isatty() and not self.shutdown_pending.is_set():
            self.warm_shutdown()
        else:
            self.cold_shutdown()
        logger.debug("Handled SIGINT")

    def _handle_sigterm(self, signum, frame):
        """Initiates a warm shutdown."""
        logger.warning("Catched SIGTERM")
        self.warm_shutdown()

    def _handle_sigquit(self, signum, frame):
        """Drop the current task and exit."""
        logger.warning("Catched SIGQUIT")
        if self._current_message:
            try:
                logger.warning("Dropping current task...")
                self._current_message.channel.basic_reject(
                    self._current_message.delivery_tag, requeue=False)
                self._current_message.channel.close()
            except Exception:
                logger.critical("Cannot send ACK for the current task.")
                traceback.print_exc()
        logger.warning("Exiting...")
        _exit(0)


def _exit(code):
    sys.stdout.flush()
    sys.stderr.flush()
    os._exit(code)


def print_stack(sig, frame):
    print('=' * 70)
    print(''.join(traceback.format_stack()))
    print('-' * 70)



# coding=utf8
import os
import re
from setuptools import setup


def read(*fname):
    with open(os.path.join(os.path.dirname(__file__), *fname)) as f:
        return f.read()


def get_version():
    for line in read('kuyruk', '__init__.py').splitlines():
        m = re.match(r"""__version__\s*=\s*['"](.*)['"]""", line)
        if m:
            return m.groups()[0].strip()
    raise Exception('Cannot find version')


setup(
    name='Kuyruk',
    version=get_version(),
    author=u'Cenk Altı',
    author_email='cenkalti@gmail.com',
    keywords='rabbitmq distributed task queue',
    url='https://github.com/cenkalti/kuyruk',
    packages=['kuyruk'],
    include_package_data=True,
    install_requires=[
        'amqp>=1.4.6',
        'setproctitle>=1.1.8',
        'blinker>=1.3',
        'argparse>=1.2.1',
    ],
    description='Simple task queue',
    long_description=read('README.rst'),
    zip_safe=True,
    entry_points={
        'console_scripts': [
            'kuyruk = kuyruk.__main__:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Object Brokering',
        'Topic :: System :: Distributed Computing',
    ],
)



from blinker import Signal

#: Sent when the task decorator is applied.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
task_init = Signal()

#: Sent before the task is applied.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
task_preapply = Signal()

#: Sent after the task is applied.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
task_postapply = Signal()

#: Sent before the wrapped function is executed.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
task_prerun = Signal()

#: Sent after the wrapped function is executed.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
task_postrun = Signal()

#: Sent when the wrapped function is returned.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
task_success = Signal()

#: Sent when the wrapped function raises an exception.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
#:    * exc_info: Return value of ``sys.exc_info()``
task_error = Signal()

#: Sent when the task fails after all retries(if any).
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
#:    * exc_info: Return value of ``sys.exc_info()``
task_failure = Signal()

#: Sent before the task is sent to queue.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
#:    * description: dict representation of the task
task_presend = Signal()

#: Sent after the task is sent to queue.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
#:    * description: dict representation of the task
task_postsend = Signal()

#: Sent when the task fails.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * worker: The Worker object
#:    * task: Task object
#:    * args: Positional arguments to the task
#:    * kwargs: Keyword arguments to the task
#:    * description: dict representation of the task
#:    * exc_info: Return value of ``sys.exc_info()``
worker_failure = Signal()

#: Sent when the worker is initialized.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * worker: The Worker object
worker_init = Signal()

#: Sent when the worker shuts down.
#:
#: Arguments:
#:    * sender: Kuyruk object
#:    * worker: The Worker object
worker_shutdown = Signal()
